/*
   Requetes 2MASS dans catalogues binaires.

   01/02/2002 : - requete par identificateur: parametre -i HHMMSSs+DDMMSS 

   20/11/2001 : - ajout du signe de la declinaison (disparu???)
                - modif de l'aide visible (suppression options -debug, -verb)

   05/03/2001 : default radius = 2.5 arcmin
   Reecriture du main avec decoupage en subroutines
   Correction du bug pour max 1 objet

   ------------
   Janvier 2001.
   Lecture du bit marquant les identificateurs multiples.
   ------------
     Nouveaux Headers (plus compacts).
     Decodage automatise dans chaine binaire (sDecode)
     Chaine de commande debuguee (modele pour les autres programmes!)
     Parcours optimise dans les repertoires et fichiers 
     Gestion de swap eventuel pour les clones Vizier sous Linux (headers)

     root_2MASS (variable d'environnement) repertoire racine des fichiers data

     v4.1: version stable et rapide (~21 microsecondes par source)
     ATTENTION: fseek() est extremement lent!!!! Il vaut mieux utiliser 
                un ou plusieurs fread() a la place. fseek() peut doubler
                le temps total (la conso de CPU systeme explose!).

     v4.2:
     Flag -W enleve (pas de sortie avec tags HTML)
     v4.3:
     if (DEBUG) enleves dans les boucles les + souvent executees
     v4.4:
     test swap bytes pour portage sur PC au CADC

*/


#include "co2mass_2.h"

static char *theProg;
static char usage[] = "[-h] [-r[sd] radius] [-c center] [-m max_records] \n\
[-f input_file] [-i HHMMSSs+DDMMSS] [-root data_directory] [-debug] [-verb]\n\
";

static char help[] = "\
-h: display this help\n\
-r: target radius in arcmin ; -rs = target radius in arcsec\n\
-rd = target radius in degrees\n\
-c: target center in decimal \n\
-m: max number of stars to retrieve \n\
-f: input file. use '-' for stdin \n\
-i: query by identifier \n\
-root: root directory for data files (default is $root_2MASS)\n\
";

static double _factor_ = 1;
static double _factor2_ = 1;
static int swapping; 
static int byId = 0;
static char theId[14];

void check_swap();
static void swap1 (short *, int);
static void swapl (long *, int);
static void makeCone(Cone *);
void findTheRoot(char *);
void readTheData(enregistrement *, FILE *);
void readNSHead(FILE*, Frame *);
struct noeud * elagage(struct noeud *);
struct noeud * loopInFrame(FILE*, double, double, Cone, long, struct noeud *);
struct noeud * loopInFile(double, double, Cone, double, char [], FILE*, FILE*, struct noeud *);
struct noeud * loopBase(struct noeud *, Cone, char *);
static int get_center (char *, double [2]);
static int match_num (char *, double *);
static int match_unum (char *, double *);

/************************************************************************/
/* boucle sur les fichiers de la base                                   */
/************************************************************************/
struct noeud * loopBase(struct noeud *racine, Cone cone, char *rootDir) {

  double ra, dec, theRA;
  char decDir[FILENAME_MAX];
  char binFile[FILENAME_MAX], hbFile[FILENAME_MAX];
  FILE *filHB, *filBIN;
  struct stat *filstat;
  char flagFrame[(NFRAMES_RA*NFRAMES_DEC)];
 
  filstat = malloc(sizeof(struct stat));

  /* START DECODING */
  for (dec = (CUT_DEC*floor(cone.delinf/CUT_DEC)); 
       dec < CUT_DEC*ceil(cone.delsup/CUT_DEC); dec+=CUT_DEC) {

    sprintf(decDir, "%s/dec%+03d", rootDir, (int)dec);

    if ( stat(decDir, filstat) != -1 ) {

      for (ra = (CUT_RA*floor(cone.alinf/CUT_RA)); 
           ra < CUT_RA*ceil(cone.alsup/CUT_RA); ra+=CUT_RA) {

        theRA = ra;
        if (ra < 0.0) theRA += 360;
        if (ra >= 360.0) theRA -= 360;

        sprintf(binFile, "%s/ra%03d.bin", decDir, (int)theRA);
        sprintf(hbFile, "%s/ra%03d.hb2", decDir, (int)theRA);
             
        if ( stat(binFile, filstat) != -1 ) {

          filBIN = fopen(binFile,"rb"); filHB = fopen(hbFile,"rb");  

          fread(&flagFrame, sizeof(char), NFRAMES_RA*NFRAMES_DEC, filHB);
          racine = loopInFile(ra, dec, cone, theRA, flagFrame, filHB, filBIN, racine);

          fclose(filBIN); fclose(filHB);

        } /* end if stat(binFile) */
      } /*  end for ra */
    } /* end if stat(dirDec) */
  } /* end for dec */
  
  return(racine);
}
/************************************************************************/
/* boucle sur les frames dans un fichier                                */
/************************************************************************/
struct noeud * loopInFile(double ra, double dec, Cone cone, double theRA, char flagFrame[], 
			  FILE* filHB, FILE* filBIN, struct noeud *racine) {

  FileBound limits;
  int fra, fdec; 
  Frame f; 

  for (fdec=0; fdec<NFRAMES_DEC; fdec++) {
    
    limits.alinf = ra;
    limits.alsup = ra + CUT_RA;
    limits.delinf = dec + fdec*WFRAME_DEC;
    limits.delsup = dec + (fdec+1)*WFRAME_DEC;

    if ( isFileOnCone(limits, cone) ) {
      
      for (fra=0; fra<NFRAMES_RA; fra++) {

        if (flagFrame[fdec*NFRAMES_RA+fra]) /* READ header of frame */
          readNSHead(filHB, &f);
        else
          f.nbElt = f.size = 0;

        f.alinf = ra + fra*WFRAME_RA;
        f.delinf = limits.delinf;               
        f.alsup = f.alinf + WFRAME_RA;
        f.delsup = f.delinf + WFRAME_DEC;  

        if ( isFrameOnCone(f, cone) && f.nbElt)
          racine = loopInFrame(filBIN, theRA+fra*WFRAME_RA, f.delinf, 
                               cone, f.nbElt, racine);
        else
          fseek(filBIN, f.size, SEEK_CUR);
       
      } /* end for fra */
    } /* end if is on (dec frame) */
    else {
      for (fra=0; fra<NFRAMES_RA; fra++) {
        if (flagFrame[fdec*NFRAMES_RA+fra]) {
          readNSHead(filHB, &f);
          fseek(filBIN, f.size, SEEK_CUR);
        }
      }
    }
  } /* end for fdec */

  return(racine);

}
/************************************************************************/
/* boucle sur les elements d'une frame concernee par une requete        */
/************************************************************************/
struct noeud * loopInFrame(FILE *filBIN, double aoff, double doff, Cone cone, 
                 long n, struct noeud *racine) {

  unsigned char binPOS[LGB2P];
  enregistrement rec;
  double dst;  
  char buffer[(3*(LGB2M+1)+LGB2P)];
  long i;
  char thisId[14];

  for (i=1; i<=n; i++) {

    fread(&binPOS, sizeof(char), LGB2P, filBIN);
    rec.pos = posDecode(binPOS, aoff, doff);

    dst = sphDst(cone.raC, cone.decC, rec.pos.ra, rec.pos.dec);

    tested++; 

    if ( dst <= cone.r) {

      if (!byId && ++matched > mrec) truncated++; /* search by position */
      if (byId) {                                 /* search by identificator */
   	 strcpy(thisId, makeName(rec.pos));
         if (!strcmp(thisId,theId)) { /* good id: it's a match */
             if (++matched > mrec) truncated++;
         }
      }

      if (matched == mrec+1)  /* trouver le max au depart */
        maxR = maxArbre(racine);

      /* on ajoute un enregistrement dans l'arbre */
      if ( dst < maxR || !truncated ) {
        readTheData(&rec, filBIN);
                                                               
        racine = ajoutarbre(racine, rec, dst);                
        
        if (truncated) { /* Si troncated: elaguer + gde feuille */
          racine = elagage(racine);
          maxR = maxArbre(racine);
        }                             
      } else /*  if not ( dst < maxR || !truncated ) */
        fread(&buffer, sizeof(char), 
              (LGB2O*rec.pos.isopt +3*LGB2M 
               +rec.pos.psfbiJ +rec.pos.psfbiH 
               +rec.pos.psfbiK), filBIN);

    } else /* END if ( dst <= cone.r) */
      fread(&buffer, sizeof(char), 
            (LGB2O*rec.pos.isopt +3*LGB2M 
             +rec.pos.psfbiJ +rec.pos.psfbiH 
             +rec.pos.psfbiK), filBIN);       
  } /*  END for (i=1; i<=f.nbElt; i++) */

  return(racine);

}
/************************************************************************/
/* lecture headers                                                      */
/************************************************************************/
void readNSHead(FILE* filHB, Frame *f) {
  fread(&(f->nbElt), sizeof(long), 1, filHB);
  fread(&(f->size), sizeof(long), 1, filHB);
  if (swapping) {
    swapl(&(f->nbElt), 1);
    swapl(&(f->size), 1);
  }
}
/************************************************************************/
/* lecture du reste des donnees                                         */
/************************************************************************/
void readTheData(enregistrement *rec, FILE *filBIN) {

  unsigned char binMAG[(LGB2M+1)], binOPT[LGB2O];

  if (rec->pos.isopt) {
    fread(&binOPT, sizeof(char), LGB2O, filBIN);
    rec->opt = optDecode(binOPT);
  } else rec->opt.n_mchs = 0;
                        
  fread(&binMAG, sizeof(char), 
        (rec->pos.psfbiJ ? LGB2M+1 : LGB2M), filBIN);
  rec->J = magDecode(binMAG, rec->pos.psfbiJ);
  rec->pos.mult_id = sDecode(binMAG,37,1);

  fread(&binMAG, sizeof(char), 
        (rec->pos.psfbiH ? LGB2M+1 : LGB2M), filBIN);
  rec->H = magDecode(binMAG, rec->pos.psfbiH);

  fread(&binMAG, sizeof(char), 
        (rec->pos.psfbiK ? LGB2M+1 : LGB2M), filBIN);
  rec->K = magDecode(binMAG, rec->pos.psfbiK);
}
/************************************************************************/
/* elagage de l'arbre binaire                                           */
/************************************************************************/
struct noeud * elagage(struct noeud *racine) {

  struct noeud *spy, *elag;

  spy = allouernoeud();
  elag = allouernoeud();
  if (racine->droit == NULL) { /* elaguer racine! */
    elag = racine;
    racine = racine->gauche;
    free(elag);
  } else {
    spy = racine;
    while (spy->droit->droit != NULL)
      spy = spy->droit;
    elag = spy->droit;
    spy->droit = spy->droit->gauche;
    free(elag);
  }

  return(racine);
}

/************************************************************************/
/* 2MASS root directory                                                 */
/************************************************************************/
void findTheRoot(char *rootDir) {

  struct stat filstat;
  char fic[FILENAME_MAX];
  char *p;

  if (strcmp(rootDir, "") == 0) {
    if (p = getenv("root_2MASS"))
      strcpy(rootDir, p) ;
    else if (*theProg == '/') { /* Use the Program Path */
      for (p=theProg+strlen(theProg)-1; *p != '/'; p--) ;
      if (strncmp(p-4, "/bin/", 4) == 0) {
        strncpy(rootDir, theProg, (p-theProg)) ;
        p = rootDir + (p-theProg) ;
        strcpy(p-4, "/data") ;
      }
    }
    if (strcmp(rootDir, "") == 0) 
      strcpy(rootDir, "/m/5/2m1/data"); 
  }

  strcpy(fic, rootDir);
  strcat(fic,"/dec-18/ra000.bin");
  if (stat(fic, &filstat) == -1) {   
    fprintf(stderr, "Empty root directory %s (datafiles not found)\n", rootDir); 
    exit(1);
  }

  if (DEBUG) fprintf(stdout, "Root directory: [%s]\n", rootDir);

}
/************************************************************************/
/* define limits of the search cone                                     */
/************************************************************************/
static void makeCone(Cone *cone) {

  double arcra, dra;
  
  if (cone->r > 10.0)  /* Eviter les requetes trop grosses ???? */
    {
      fprintf(stdout,"# Search radius truncated to 10.0 deg !!!\n");
      cone->r = 10.0;
    }

  /***************************************/
  /* calcul des extremites de la requete */
  cone->delinf = cone->decC - cone->r;
  if (cone->delinf < -90.) 
    cone->delinf = -90.;
  cone->delsup = cone->decC + cone->r;
  if (cone->delsup > 90.) 
    cone->delsup = 90.;

  cone->alinf = 0.;
  cone->alsup = 360.;
  arcra = (sind(cone->r))/(cosd(cone->decC));
  if (arcra < 1.)
    {
      dra = asind(arcra);
      if (cone->r > 90.)
        dra = 180. - dra;
      cone->alinf = cone->raC - dra;
      /* if(cone->alinf < 0.) cone->alinf +=360.; */
      cone->alsup = cone->raC + dra;
      /* if(cone->alsup > 360.) cone->alsup -=360.; */
    }
  /***************************************/

  if (DEBUG) {
    fprintf(stdout, "Cone centre sur %f %f\n",
            cone->raC, cone->decC);
    fprintf(stdout, "Rayon de recherche: %f\n", cone->r);
    fprintf(stdout, "Intervalle: [ra] %f %f [dec] %f %f\n",
            cone->alinf, cone->alsup, cone->delinf, cone->delsup);
  }

}

/************************************************************************
 #    #    ##       #    #    #
 ##  ##   #  #      #    ##   #
 # ## #  #    #     #    # #  #
 #    #  ######     #    #  # #
 #    #  #    #     #    #   ##
 #    #  #    #     #    #    #
************************************************************************/
main (int argc, char **argv)
{
  char rootDir[FILENAME_MAX]="";
  char argline[BUFSIZ];
  Cone cone;

  struct stat *filstat;
  struct noeud *racine;

  char *arg, p;
  int fromfile=0;
  FILE* fin;
  char finName[FILENAME_MAX];
  char line[BUFSIZ];
  double coo[2] ;
  int i;

  /* default radius = 2.5 arcmin */
  cone.r = 2.5/60.0;

  filstat = malloc(sizeof(struct stat));

  racine = allouernoeud();

  theProg = argv[0];
  /************************************************/
  if (argc < 2) 
    {
      fprintf(stderr, "Usage: %s %s\n%s\n", theProg, usage, help);
      exit(1);
    }

  while (--argc >= 0 && (arg = *++argv)!=NULL) 
  {
    if (arg[0]=='-' && tolower(arg[1])=='h')
      {
        fprintf(stderr, "Usage: %s %s%s", theProg, usage, help) ;
        exit(0) ;
      }

    if (!strcmp(arg, "-root"))  /* root directory */
      {
        if ((*++argv)==NULL) { 
          fprintf(stderr, "Error in param -root\n"); 
          exit(1);
        }
        strcpy(rootDir, *argv); 
        if (stat(rootDir, filstat) == -1) {
          fprintf(stderr, "Invalid root directory %s\n", rootDir); 
          exit(1);
        }
        strcpy(argline, rootDir);
        strcat(argline,"/dec-18/ra000.bin");
        if (stat(argline, filstat) == -1) {       
          fprintf(stderr, "Empty root directory %s (datafiles not found)\n", rootDir); 
          exit(1);
        }
        argc--;
        continue ;
      }

    if (strstr(arg, "-r")) {  /* Next parameter = Radius */
        if (*++argv==NULL)
          { fprintf(stderr,"Error in param -r\n"); 
            exit(1);}
            switch(arg[2]) {
            case 'd': _factor_ = 1; break ;
            case 's': _factor_ = 1./3600. ; break ;
            default:  _factor_ = 1./60. ; break ;
            }
          cone.r = atof(*argv) * _factor_;
          argc--; 
        continue;
       }

   if (strstr(arg, "-i")) {  /* Next parameter = identificateur */
        if (*++argv==NULL)
          { fprintf(stderr,"Error in param -i\n"); 
            exit(1);}
        if (strlen(*argv)!=14)
          { fprintf(stderr,"Bad identifier. Should be formatted as HHMMSSs+DDMMSS\n"); 
            exit(1);}
        byId = 1;
	strcpy(theId, *argv);
        cone.r = 3./3600.; /* default: 3 arcsec */
	/* id is HHMMSSs+DDMMSS */
	p=theId[0];              /* extract RA from string */
        cone.raC = 10*atof(&p);
	p=theId[1];
        cone.raC += atof(&p);
	p=theId[2];
        cone.raC += atof(&p) / 6.;
	p=theId[3];
        cone.raC += atof(&p) / 60.;
	p=theId[4];
        cone.raC += atof(&p) / 360.;
	p=theId[5];
        cone.raC += atof(&p) / 3600.;
	p=theId[6];
        cone.raC += atof(&p) / 36000.;
	cone.raC *= 15;  /* hours -> deg */
	p=theId[8];
	cone.decC = 10*atof(&p);    /* extract DEC from string */
	p=theId[9];
        cone.decC += atof(&p);
	p=theId[10];
        cone.decC += atof(&p) / 6.;
	p=theId[11];
        cone.decC += atof(&p) / 60.;
	p=theId[12];
        cone.decC += atof(&p) / 360.;
	p=theId[13];
        cone.decC += atof(&p) / 3600.;
	p=theId[7];                   /* apply sign */
	if (p == '-') cone.decC = -cone.decC;

	if (mrec == 100)
	   mrec = 5; /* if not changed from default, set to 5
                         never more than 5 objects with same id */

        argc--;
        continue;
      }

    if (!strcmp(arg, "-m"))   /* Max number of records */
      {
        if (*++argv==NULL)
          { fprintf(stderr,"Error in param -m\n"); 
            exit(1);}
         mrec = atoi(*argv) ;
         if (mrec < 1) mrec = 1 ;
         argc-- ;
         continue;
      }


    if (!strcmp(arg, "-c"))   /* Get Center */
      {
        if (*++argv==NULL)
          { fprintf(stderr,"Error in param -c\n"); 
          exit(1);}         

         if (!strchr(*argv,'+') && !strchr(*argv,'-'))
           {
              /* c'est en 2 parties */
              fflush(stdout);
              cone.raC = atof(*argv);
              argc-- ;

              if (*++argv==NULL)
                { fprintf(stderr,"Error in param -c\n");  
                exit(1);}
              cone.decC = atof(*argv);
              argc-- ;
           }
         else
           {
              /* tout en 1 bloc */
              fflush(stdout);
              if (strchr(*argv,'+'))   
                { cone.decC = atof(strchr(*argv,'+'));
                  strncpy(argline,*argv,strchr(*argv,'+')-*argv);
                  cone.raC = atof(argline);
                }
              else  
                { cone.decC = atof(strchr(*argv,'-'));
                  strncpy(argline,*argv,strchr(*argv,'-')-*argv);
                  cone.raC = atof(argline);
                }
              argc-- ;
           }
         continue ;
      }

    /* positions in input file */
    if (!strcmp(arg, "-f"))
      {
        fromfile = 1;
        if ((*++argv)==NULL) { 
          fprintf(stderr, "Error in param -f\n"); 
          exit(1);
        }
        strcpy(finName, *argv);
        if (!strcmp(finName,"-"))
          fin = stdin;
        else 
          if ((fin = fopen(finName, "r")) == NULL) 
            { fprintf(stderr, "Can't open file %s\n", finName); 
              exit(1); 
            } 
        argc--;
        continue;
      }

    /* DEBUG: Prints everything */
    if (!strcmp(arg, "-debug"))
      {
        DEBUG = 1;
        continue;
      }

    /* VERBOSE Prints more */
    if (!strcmp(arg, "-verb"))
      {
        VERB = 1;
        continue;
      }

    fprintf(stderr, "invalid option %s\n", arg);
    exit(1);

  }
  /**####################################################################**/
  /**####################################################################**/


  if (DEBUG) fprintf(stdout, "Command line read\n");
  fflush(stdout);

  if (DEBUG) {
    fprintf(stdout, "sizeof(char)=%d\nsizeof(int)=%d\nsizeof(long)=%d\n", 
            sizeof(char), sizeof(int), sizeof(long));
  }
  /* byte swap if PC CADC */
  check_swap();

  findTheRoot(rootDir);
  truncated = matched = tested = 0;

  if (fromfile) {
    do {
      fgets(line, BUFSIZ, fin);
      if (feof(fin)) break;
      for (i=strlen(line)-1; i>=0; i--)
        if (isspace(line[i]))
          line[i] = '\0';
        else
          break;

      if (line[0]=='#' || line[0]=='\0') /* recopie commentaire */

        fprintf(stdout, "#%s\n", line); 
 
      else { /* requete position */

         get_center(line, coo) ;
         cone.raC = coo[0];
         cone.decC = coo[1];

         fprintf(stdout, "# %s [Within %7.1f arcsec of %010.6f%+010.6f]\n", 
                 line, 3600*cone.r, cone.raC, cone.decC);
         racine = NULL;
         makeCone(&cone);  
         racine = loopBase(racine, cone, rootDir);
         affarbre(racine); 
         if (truncated)
           fprintf(stdout,"#+++ Truncated to %ld of %ld matches (%ld tested)\n", 
                          mrec, matched, tested);
         else
           fprintf(stdout,"#--- %ld matches (%ld tested)\n", matched, tested);
         truncated = matched = tested = 0;
      }
    } while (!feof(fin));

  } else { /* normal request */

    racine = NULL;   /* Initialisation du premier noeud */
    makeCone(&cone);  /* compute search cone */
    racine = loopBase(racine, cone, rootDir); /* browse database */
    affarbre(racine); /* print result */
    if (truncated)
      fprintf(stdout,"#+++ Truncated to %ld of %ld matches (%ld tested)\n", 
                     mrec, matched, tested);
    else
      fprintf(stdout,"#--- %ld matches (%ld tested)\n", matched, tested);
  }

  return(0);

}  /* END of main */ 


/****************************************************************************/
void printEnr(enregistrement elt, FILE *f)
{
  int placeholder = 0;
  char empty[10];
  char thisId[14];
  
  strcpy(thisId, makeName(elt.pos));
  if (byId && strcmp(thisId,theId)) { /* search with -i, and wrong id: do not print */
	return; 
  }

  /* la ligne suivante permet de mettre "null" dans les
     champs vides au lieu de blancs (recompiler et verifier format) */
  strcpy(empty,"");

  if (elt.J.rd_flg == 8 && elt.H.rd_flg == 8 && elt.K.rd_flg == 8) 
    placeholder = 1;

  fprintf(f, "%10.6f ", elt.pos.ra);
  fprintf(f, "%+10.6f ", elt.pos.dec);

  if (placeholder) 
    fprintf(f, "%4s %4s %4s ", empty, empty, empty); 
  else 
    {
      if (fabs(elt.pos.err_maj - 2.55) < 5e-3) 
        fprintf(f, "%4s ", empty);
      else
        fprintf(f, "%4.2f ", elt.pos.err_maj);

      if (fabs(elt.pos.err_min - 2.55) < 5e-3) 
        fprintf(f, "%4s ", empty);
      else
        fprintf(f, "%4.2f ", elt.pos.err_min);

      if (elt.pos.err_ang == 511) 
        fprintf(f, "%4s ", empty);
      else
        fprintf(f, "%4d ", elt.pos.err_ang);   
    }

  if (elt.pos.mult_id)
    fprintf(f, "*");
  else
    fprintf(f, " ");

  fprintf(f, "%14s ", thisId);

  /* ## J ## m, msig, msigcom */
  if (fabs(elt.J.m - 32.767) < 5e-4)  
    fprintf(f, "%7s ", empty);
  else 
    fprintf(f, "%7.3f ", elt.J.m);

  if (fabs(elt.J.msig - 16.383) < 5e-4) 
    fprintf(f, "%5s ", empty);
  else 
    fprintf(f, "%5.3f ", elt.J.msig);

  if (fabs(elt.J.msigcom - 0.127) < 5e-4) 
    fprintf(f, "%5s ", empty);
  else 
    fprintf(f, "%5.3f ", elt.J.msig+elt.J.msigcom);

 
  /* ## H ## m, msig, msigcom */
  if (fabs(elt.H.m - 32.767) < 5e-4)  
    fprintf(f, "%7s ", empty);
  else 
    fprintf(f, "%7.3f ", elt.H.m);

  if (fabs(elt.H.msig - 16.383) < 5e-4) 
    fprintf(f, "%5s ", empty);
  else 
    fprintf(f, "%5.3f ", elt.H.msig);

  if (fabs(elt.H.msigcom - 0.127) < 5e-4) 
    fprintf(f, "%5s ", empty);
  else 
    fprintf(f, "%5.3f ", elt.H.msig+elt.H.msigcom);

 

  /* ## K ## m, msig, msigcom */
  if (fabs(elt.K.m - 32.767) < 5e-4)  
    fprintf(f, "%7s ", empty);
  else 
    fprintf(f, "%7.3f ", elt.K.m);

  if (fabs(elt.K.msig - 16.383) < 5e-4) 
    fprintf(f, "%5s ", empty);
  else 
    fprintf(f, "%5.3f ", elt.K.msig);

  if (fabs(elt.K.msigcom - 0.127) < 5e-4) 
    fprintf(f, "%5s ", empty);
  else 
    fprintf(f, "%5.3f ", elt.K.msig+elt.K.msigcom);

 

  /* concatenate rd_flg */
  fprintf(f, "%d%d%d ", elt.J.rd_flg, elt.H.rd_flg, elt.K.rd_flg);

  /* concatenate bl_flg */
  if (elt.J.bl_flg == 7)
    fprintf(f, "%4s ", empty);
  else
    {
      fprintf(f, " %d%d%d ", elt.J.bl_flg, elt.H.bl_flg, elt.K.bl_flg);
    }

  /* concatenate cc_flg */
  if (elt.J.cc_flg == 'X')
    fprintf(f, "%4s ", empty);
  else
    {
      fprintf(f, " %c%c%c ", elt.J.cc_flg, elt.H.cc_flg, elt.K.cc_flg);
    }

  /* extd_flg */
  if (elt.pos.extd_flg == 3)
    fprintf(f, "%4s ", empty);
  else
    fprintf(f, "   %d ", elt.pos.extd_flg);

  /* mp_flg ambiguous???? */
  fprintf(f, "%d ", elt.pos.mp_flg);

  /* optical data */
  if (elt.pos.isopt && !placeholder)
    {
     fprintf(f, "   %c ", elt.opt.id);

     if (fabs(elt.opt.b_m - 40.95) < 5e-3) 
       fprintf(f, "%5s ", empty);
     else if (fabs(elt.opt.b_m - 40.9) < 5e-3) 
       fprintf(f, "99.90 ");
     else      
       fprintf(f, "%5.2f ", elt.opt.b_m);

     if (fabs(elt.opt.r_m - 40.95) < 5e-3) 
       fprintf(f, "%5s ", empty);
     else if (fabs(elt.opt.r_m - 40.9) < 5e-3) 
       fprintf(f, "99.90 ");
     else 
       fprintf(f, "%5.2f ", elt.opt.r_m);

     if (fabs(elt.opt.dist - 5.11) < 5e-3) 
       fprintf(f, "%5s ", empty);
     else 
       fprintf(f, "%5.2f ", elt.opt.dist);
     
     if (elt.opt.phi == 511) 
       fprintf(f, "%4s ", empty);
     else 
       fprintf(f, "%4d ", elt.opt.phi);
    }
  else
    fprintf(f, "%4s %5s %5s %5s %4s ",
                empty, empty, empty, empty, empty);

  if (elt.opt.n_mchs == 31) 
    fprintf(f, "%4s ", empty);
  else 
    fprintf(f, "%4d ", elt.opt.n_mchs);

  /* date hemis scan */
  if (placeholder) 
    fprintf(f, "%11s %4s %4s ", empty, empty, empty);
  else 
    {
      if (!strcmp(elt.pos.date, " 01/01/1997"))
        fprintf(f, "%11s ", empty);
      else
        fprintf(f, "%11s ", elt.pos.date);

      fprintf(f, "   %c ", elt.pos.hemis);
      
      if (elt.pos.scan == 255)
        fprintf(f, "%4s ", empty);
      else
        fprintf(f, "%4d ", elt.pos.scan);
    }

  /* id glon glat */
  if (elt.pos.id == 1048575)
    fprintf(f, "%6s ", empty);
  else
    fprintf(f, "%6ld ", elt.pos.id);

  if (placeholder) /* round values to 1st decimal ! */
    {
      elt.pos.glon = rint(10.*elt.pos.glon)/10.;
      elt.pos.glat = rint(10.*elt.pos.glat)/10.;
    }
  fprintf(f, "%10.6f ", elt.pos.glon);
  fprintf(f, "%10.6f ", elt.pos.glat);
  
  /* psfchi */
  if (fabs(elt.J.psfchi + 0.01) < 5e-3) 
    fprintf(f, "%6s ", empty);
  else 
    fprintf(f, "%6.2f ", elt.J.psfchi);

  if (fabs(elt.H.psfchi + 0.01) < 5e-3) 
    fprintf(f, "%6s ", empty);
  else 
    fprintf(f, "%6.2f ", elt.H.psfchi);

  if (fabs(elt.K.psfchi + 0.01) < 5e-3) 
    fprintf(f, "%6s ", empty);
  else 
    fprintf(f, "%6.2f ", elt.K.psfchi);

  /* m_stdap */
  if (fabs(elt.J.m_stdap - 131.050) < 5e-4)
    elt.J.m_stdap = -99.999;
  if (fabs(elt.J.m_stdap - 131.071) < 5e-4) 
    fprintf(f, "%7s ", empty);
  else 
    fprintf(f, "%7.3f ", elt.J.m_stdap);
  if (fabs(elt.J.msig_stdap - 16.383) < 5e-4)
    fprintf(f, "%6s ", empty);
  else 
    fprintf(f, "%6.3f ", elt.J.msig_stdap);

  if (fabs(elt.H.m_stdap - 131.050) < 5e-4)
    elt.H.m_stdap = -99.999;
  if (fabs(elt.H.m_stdap - 131.071) < 5e-4) 
    fprintf(f, "%7s ", empty);
  else 
    fprintf(f, "%7.3f ", elt.H.m_stdap);
  if (fabs(elt.H.msig_stdap - 16.383) < 5e-4) 
    fprintf(f, "%6s ", empty);
  else 
    fprintf(f, "%6.3f ", elt.H.msig_stdap);

  if (fabs(elt.K.m_stdap - 131.050) < 5e-4)
    elt.K.m_stdap = -99.999;
  if (fabs(elt.K.m_stdap - 131.071) < 5e-4) 
    fprintf(f, "%7s ", empty);
  else 
    fprintf(f, "%7.3f ", elt.K.m_stdap);
  if (fabs(elt.K.msig_stdap - 16.383) < 5e-4) 
    fprintf(f, "%6s ", empty);
  else 
    fprintf(f, "%6.3f ", elt.K.msig_stdap);

  /* ndet_flg */
  if (placeholder) 
    fprintf(f, "%8s", empty);
  else /* 7 can't be used for null cause it happens !!! */
    {
      fprintf(f, "  %d%d%d%d%d%d", 
              elt.J.ndet_flg1, elt.J.ndet_flg2, elt.H.ndet_flg1, 
              elt.H.ndet_flg2, elt.K.ndet_flg1, elt.K.ndet_flg2);
    }
  fprintf(f, "\n");

}

/******************************************************************************/
optical optDecode(unsigned char *binstr)
{
  long long nb;
  int lgb, btot=1;
  optical elt;
                
  /* n_mchs */

  lgb = 5;
  elt.n_mchs = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* id_opt */

  lgb = 1;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;
  elt.id = (nb == 0) ? 'T' : 'U';

  /* dist */

  lgb = 9;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;
  elt.dist = 1.0e-2*nb;

  /* phi */

  lgb = 9;
  elt.phi = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* b_m */

  lgb = 12;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  if (nb > 2500 && nb <= 4000) nb -= 4000;
  elt.b_m = 1.0e-2*nb;

  /* r_m */

  lgb = 12;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  if (nb > 2500 && nb <= 4000) nb -= 4000;
  elt.r_m = 1.0e-2*nb;

  return(elt);
}

/******************************************************************************/

databand magDecode(unsigned char *binstr, int bimode)
{
  long long nb;
  int lgb, btot=1;
  databand elt;
  char ccstring[] = "0PDSRCBMX";

  /* m */
 
  lgb = 15;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  if (nb > 25000 && nb <= 32000) nb -= 32000;
  elt.m = 1.0e-3*nb;

  if (fabs(elt.m - 25.) < 5e-4) /* placeholder or saturated */
    elt.m = -99.999;

  /* msig */
 
  lgb = 14;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;
  elt.msig = 1.0e-3*nb;

  /* msigcom */
 
  lgb = 7;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;
  elt.msigcom = 1.0e-3*nb;

  /* Trou dans le codage !!!! */
  btot +=4;

  /* rd_flg */
 
  lgb = 3;
  elt.rd_flg = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  if (elt.rd_flg == 6) /* cause that's the way it's coded ! */
    elt.rd_flg = 8; 

  /* bl_flg */
 
  lgb = 3;
  elt.bl_flg = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* cc_flg */
 
  lgb = 4;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  strncpy(&(elt.cc_flg), &(ccstring[nb]), 1);

  /* m_stdap */
 
  lgb = 17;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  if (nb > 100000 && nb <= 130000) nb -= 130000;
  elt.m_stdap = 1.0e-3*nb;

  /* msig_stdap */
 
  lgb = 14;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  elt.msig_stdap = 1.0e-3*nb;

  /* ndet_flg */
 
  lgb = 3;
  elt.ndet_flg1 = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  lgb = 3;
  elt.ndet_flg2 = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* psfchi */
 
  if (bimode) {
    lgb = 17;
    nb = sDecode(binstr,btot,lgb); 
    btot +=lgb;
  } else {
    lgb = 9;
    nb = sDecode(binstr,btot,lgb); 
    btot +=lgb;
  }

  elt.psfchi = 1.0e-2*(nb-1); /* mind the offset ! */

  return(elt);

}

/******************************************************************************/


position posDecode(unsigned char *binstr, double ra, double dec)
{
  long long nb;
  int lgb, btot=1;
  position elt;

  /* ra */

/*   lgb = 16; */
/*   nb = sDecode(binstr,btot,lgb);  */
/*   btot +=lgb;  */

  nb = binstr[0];            
  nb = (nb << 8) | binstr[1];
  elt.ra = nb*1.0e-6 + ra;

  /* dec */

/*   lgb = 16; */
/*   nb = sDecode(binstr,btot,lgb);  */
/*   btot +=lgb; */

  nb = binstr[2];            
  nb = (nb << 8) | binstr[3];
  elt.dec = nb*1.0e-6 + dec;
  
  eq2gal(elt.ra, elt.dec, &(elt.glon), &(elt.glat));

  /* err_maj */

/*   lgb = 8; */
/*   nb = sDecode(binstr,btot,lgb);  */
/*   btot +=lgb; */

  nb = binstr[4];            
  elt.err_maj = nb*1.0e-2;

  /* err_min */

/*   lgb = 8; */
/*   nb = sDecode(binstr,btot,lgb);  */
/*   btot +=lgb; */

  nb = binstr[5];            
  elt.err_min = nb*1.0e-2;

  /* err_ang */

  btot +=48; /* valeurs deja lues avant 6 octets ! */
  lgb = 9;
  elt.err_ang = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* extd_flg */

  lgb = 2;
  elt.extd_flg = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* mp_flg */

  lgb = 1;
  elt.mp_flg = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* isopt */

  lgb = 1;
  elt.isopt = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* date */

  lgb = 11;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  strcpy(elt.date, jd2date(nb + JDoffset));
  elt.date[10] = '\0';

  /* scan */

  lgb = 9;
  elt.scan = sDecode(binstr,btot,lgb); 
  btot +=lgb;
  
  /* hemis */

  lgb = 1;
  nb = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  if (nb == 0)
    elt.hemis = 'n';
  else
    elt.hemis = 's';

  /* id */

  lgb = 19;
  elt.id = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  /* psfbi */

  lgb = 1;
  elt.psfbiJ = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  lgb = 1;
  elt.psfbiH = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  lgb = 1;
  elt.psfbiK = sDecode(binstr,btot,lgb); 
  btot +=lgb;

  return(elt);
}

/***************************************************************/

double sphDst(double ra1,double dec1,double ra2,double dec2)
{
 double v1,v2,v3, cosdec, u[3], u0[3];

  cosdec = cosd(dec2);
  u[0] = cosdec * cosd(ra2);
  u[1] = cosdec * sind(ra2);
  u[2] = sind(dec2);

  cosdec = cosd(dec1);
  u0[0] = cosdec * cosd(ra1);
  u0[1] = cosdec * sind(ra1);
  u0[2] = sind(dec1);

  v3 = (u[0]-u0[0])*(u[0]-u0[0]) + (u[1]-u0[1])*(u[1]-u0[1]) + (u[2]-u0[2])*(u[2]-u0[2]);
  v2 = (sqrt(v3))/2.;

  v1 = 2*asind(v2);

  return(v1);

}

/*********************************************************************/
/*    AJOUTE UN NOEUD CONTENANT LE NOUVEL ENREGISTREMENT             */
/*********************************************************************/ 

struct noeud *ajoutarbre(struct noeud *p, enregistrement enr, double rayon) 
{    

   if (p==NULL) 
     {                          /* Nouveau nom. */
       p=allouernoeud();        /* Creation d'un nouveau noeud. */    
       p->rec = enr;            /* Initialisation du nouveau noeud. */
       p->r = rayon;
       p->gauche = p->droit = NULL ;
     }
   else if( rayon < p->r )                     /* r inferieur */
     p->gauche = ajoutarbre(p->gauche,enr,rayon);
   else                                        /* r superieur */
     p->droit = ajoutarbre(p->droit,enr,rayon);

   return(p);

}
          
/*********************************************************************/
/*   AFFICHE L'ARBRE DANS LA SORTIE STANDARD                         */
/*********************************************************************/

void affarbre(struct noeud *p)
{
   if (p!=NULL)
   {
     affarbre(p->gauche);

     printEnr(p->rec, stdout);  /* affiche l'enregistrement*/

     affarbre(p->droit);
   }
  
  return ;
}

/*********************************************************************/
/*  CREE UN NOEUD DANS L'ARBRE DES ENREGISTREMENTS                   */
/*********************************************************************/
struct noeud *allouernoeud()
{
  struct noeud *p;

  p = malloc(sizeof(struct noeud));
  if(p==NULL)
    {
      fprintf(stderr,"Unable to allocate memory.\n");  
      exit (1);
    } 
  return(p) ;  /* p est un pointeur */
}

/*********************************************************************/
/*  Renvoie la valeur max de r dans l'arbre                          */
/*********************************************************************/
double maxArbre(struct noeud *theSpy)
{
   while (theSpy->droit != NULL)
     theSpy = theSpy->droit;
   return(theSpy->r);
}


/*========================================================================*/
struct tm jd2tm(double JD)
/*+++
.PURPOSE Transform Julian Date to time structure
.RETURNS Time as a structure
.REMARKS  INPUT:        Julian date    
---*/
{
  register long int j, n4, nd10;
  double x;
  struct tm T; 

    x = JD + 0.5e0 + 0.5e0/86400.e0;    /* Round day    */
    j = x;
    if (x < 0)   --j; 
    x = x - j;                          /* Fraction of day      */
  
    T.tm_year = -4712;
    while (j <= 4480)   { j += 146097L; T.tm_year -= 400;}

    n4 = 4*(j+((2*((4*j-17918L)/146097L)*3)/4+1)/2-37);
    nd10=10*( ((n4-237)%1461)/4)+5;

    T.tm_year += n4/1461;
    T.tm_mon  = (nd10/306+2)%12;
    T.tm_mday = (nd10%306)/10+1;

    x *= 86400.e0;                      /* Seconds      */
    j = x;
    T.tm_hour = j/3600;
    T.tm_min  = 0;
    T.tm_sec  = j%3600;


    return(T);
}

/*========================================================================*/
char *jd2date(double JD)
/*+++
.PURPOSE Transform Julian Date to time structure
.RETURNS Time as a structure
.REMARKS INPUT: Julian date   
---*/
{
  struct tm T; 
  char *date;

  date = malloc(11*sizeof(char));

  T = jd2tm(JD);

  sprintf(date, "%02d/%02d/%04d", T.tm_mon+1, T.tm_mday, T.tm_year);
  date[10] = '\0';

  return(date);
}

/* *********************************************************************** */
char *makeName(position p)
{
  /* make ID in the form: HHMMSSs+/-DDMMSS for 2MASS sources */

  static char id[15];
  char sg;
  long bra, bdec;
  
  bra = (long)floor((p.ra+1e-7)*2400); /* tens of arcseconds (of hours) */
  sg = (p.dec<0) ? '-' : '+';
  bdec = (p.dec<0) ? (long)floor((-p.dec+1e-7)*3600) : (long)floor((p.dec+1e-7)*3600);

  sprintf(id,"%02d%02d%03d%c%02d%02d%02d", 
          bra/36000, (bra%36000)/600, bra%600, sg,
          bdec/3600, (bdec%3600)/60, bdec%60);

  return(id);
  
}

/* *********************************************************************** */
int tr_ou (double o[2] , double u[3])  
{       
  double cosdec, ra, dec;

  ra = o[0];
  dec = o[1];

  cosdec = cosd(dec);
  u[0] = cosdec * cosd(ra);
  u[1] = cosdec * sind(ra);
  u[2] = sind(dec);

  return (1);
}  

/* *********************************************************************** */
int tr_uo (double u[3], double o[2])   
{
  double r2;            /* sqrt(x*x+y*y) */

  r2 = u[0]*u[0] + u[1]*u[1]; 
  o[0] = 0.e0;
  if (r2  == 0.e0) {            /* in case of poles */
        if (u[2] == 0.e0)       return(0);
        o[1] = ( u[3]>0.e0 ? 90.e0 : -90.e0); 
        return(1); 
  }

  o[1] = atand  ( u[2] / sqrt(r2) );
  o[0]  = atan2d (u[1] , u[0]);
  if (o[0] < 0.e0) o[0] += 360.e0;
  return (1);
} 

/* *********************************************************************** */
int tr_uu(double u1[3], double u2[3], double R[3][3]) 
{
  int i,j;
  double val;
  double u_stack[3];    /* allows same address for input/output      */


  for (i=0; i<3; i++)
  {     val = 0.e0;
        for (j=0; j<3; j++)     val += R[i][j]*u1[j];
        u_stack[i] = val;
  }
  for (i=0; i<3; i++)           u2[i] = u_stack[i];     /* copies to output */
  return(1);
}

/* *********************************************************************** */
void eq2gal(double alpha, double delta, double *l, double *b)
{
  double gal_2000[3][3]={   /* J2000 to galactic rotation matrix */
        {-0.054875539726e0, -0.873437108010e0, -0.483834985808e0},
        { 0.494109453312e0, -0.444829589425e0,  0.746982251810e0},
        {-0.867666135858e0, -0.198076386122e0,  0.455983795705e0}
        };
  double u1[3],u2[3],o1[2],o2[2];

  o1[0]=alpha;
  o1[1]=delta;

  tr_ou(o1,u1);
  tr_uu(u1,u2,gal_2000);
  tr_uo(u2,o2);

  *l=o2[0];
  *b=o2[1];
}

/* *********************************************************************** */
int isIn(double value, double b1, double b2)
{
   if ( (b1<=value && value<=b2) || (b2<=value && value<=b1) )
     return(1);
   else
     return(0);
}

/* *********************************************************************** */
int isInGlob(double a, double b, double c, double d)
{
   if ( isIn(a, c, d) || isIn(b, c, d) || isIn(c, a, b) || isIn(d, a, b) )
     return(1);
   else
     return(0);
}


/* *********************************************************************** */
int isFileOnCone (FileBound fileb, Cone cone)
{
  int testAl=0;

  if (cone.alinf > cone.alsup)
    {
      testAl = isInGlob(fileb.alinf, fileb.alsup, cone.alinf, 360.) + \
        isInGlob(fileb.alinf, fileb.alsup, 0., cone.alsup);
    }
  else
    {
      testAl = isInGlob(fileb.alinf, fileb.alsup, cone.alinf, cone.alsup);
    }

  if (isInGlob(fileb.delinf, fileb.delsup, cone.delinf, cone.delsup) && testAl)
    return(1);
  else
    return(0);

}

/* *********************************************************************** */
int isFrameOnCone (Frame fra, Cone cone)
{
  int testAl=0;

  if (cone.alinf > cone.alsup)
    {
      testAl = isInGlob(fra.alinf, fra.alsup, cone.alinf, 360.) + \
        isInGlob(fra.alinf, fra.alsup, 0., cone.alsup);
    }
  else
    {
      testAl = isInGlob(fra.alinf, fra.alsup, cone.alinf, cone.alsup);
    }

  if (isInGlob(fra.delinf, fra.delsup, cone.delinf, cone.delsup) && testAl)
    return(1);
  else
    return(0);

}

/* ************************************************** */
/* Cette fonction effectue le decodage d'un long 'nb' qui
   est dans une chaine de carateres non signee 'binstr'.
   La valeur 'nb' commence au bit 'bstart' (1er bit de la chaine=1)
   et tient sur 'lgbit' bits.
   */
 
long sDecode2(unsigned char binstr[], int bstart, int lgbit)
{
  int octet;
  long nb=0 ; /* Imperatif pour mettre tous les bits a zero !! */

  /* La chaine 'binstr' a un indice de 0 pour le premier caractere, i.e.
   le premier octet. */  

  /* 'octet' est l'indice de la chaine ou demarre le segment 
     binaire a decoder*/
  octet = (int) floor ((bstart - 1)/8.0);
  
  /* on ne recupere que la partie utile de l'octet considere */
  nb = binstr[octet] & ( (1 << (8 - (bstart-1)%8) ) -1);

  /* La nouvelle longueur de la valeur a decoder 
     si le segment s'etend au dela de l'octet de depart
     (on obtient un 'lgbit' positif strictement).
     Sinon c'est la valeur du residu (on obtient un 'lgbit' negatif ou nul).*/
  lgbit -= 8 - (bstart-1)%8;
  
  while  (lgbit > 0)   {
    nb <<= 8;
    octet++;   
    nb |= binstr[octet];
    lgbit -= 8; 

  }

  /* On decale du residu vers la droite */
  nb >>= -lgbit;

 return(nb);
}
  
/****************************************************/
/*
   Effectue le decodage d'un long (nb) dans une 
   chaine de caractere non signee binstr.
   La valeur nb commence au bit bstart (1er bit de la chaine=1)
   et tient sur lgbit bits.
 */
long sDecode(unsigned char binstr[], int bstart, int lgbit)
{
  
 int k;
 int pos, bit;
 long n=0;

 k=0;
 while (k < lgbit) {
   n <<= 1;
   pos = bstart + k;
   bit = (pos%8==0) ? 0 : 8 - (pos%8);
   n |= ( (binstr[(pos-1)>>3]&(1<<bit)) >> bit );
   k++;
 }

 return(n);

}

static void swap1 (array, nshort)
/*++++++++++++++++
.PURPOSE  Swap the bytes in the array of shorts if necessary
.RETURNS  0/1/2 (type of swap)
.REMARKS  Useful for big-endian machines
-----------------*/
  short *array ;  /* IN: The array to convert   */
  int nshort ;    /* IN: The number of integers */
{
  register int m, n ;
  register short *p, *e ;
    for (p=array, e=p+nshort; p<e; p++) {
        n = *p ;
        m =  (n>>8) & 0xff ;
        m |= (n<<8) ;
        *p = m ;
    }
}

static void swapl (array, nint)
/*++++++++++++++++
.PURPOSE  Swap the bytes in the array of long if necessary
.RETURNS  0/1/2 (type of swap)
.REMARKS  Useful for big-endian machines
-----------------*/
  long *array ;  /* IN: The array to convert   */
  int nint ;    /* IN: The number of integers */
{
  char *p, *e ;
  int n ;
  
    p = (char *)array ; e = p + 4*nint ;
    if (swapping == 1) swap1((short *)array, 2*nint) ;
    else if (swapping == 2) while (p < e) {
        n = p[0] ; p[0] = p[3] ; p[3] = n ;
        n = p[1] ; p[1] = p[2] ; p[2] = n ;
        p += 4 ;
    }
}

void check_swap() {

  static int value ;
  char *v ;
  
    value = 0x010203 ;
    v = (char *)(&value) ;
    if ((v[0] == 0) && (v[1] == 1) && (v[2] == 2)) 
        swapping = 0 ;  /* No swap necessary */
    else if ((v[0] == 1) && (v[1] == 0) && (v[2] == 3)) 
        swapping = 1 ;  /* Half-word swap */
    else if ((v[0] == 3) && (v[1] == 2) && (v[2] == 1)) 
        swapping = 2 ;  /* Full-word swap */
    else {
        fprintf(stderr, "****Irrationnal Byte Swap %02x%02x%02X%02x\n", 
            v[0]&0xff, v[1]&0xff, v[2]&0xff, v[3]&0xff) ;
        exit(2) ;
    }
    if (swapping && DEBUG) fprintf(stderr, 
        "#...swapping type=%d\n", swapping) ;



}


/* Code qui interprete des coordonnees - F. Ochsenbein */

static int match_unum (char *string, double *value)
/*++++++++++++++++
.PURPOSE  Analyze the incoming and get unsigned number (no E notation)
.RETURNS  How many bytes matched
.REMARKS  Initial blanks return 0
-----------------*/
{
  char *p;
    for (p=string; isdigit(*p); p++) ;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    if (p == string) *value = 0 ;
    else *value = _factor2_ * atof(string) ;
    return(p-string);
}
static int match_num (char *string, double *value)
/*++++++++++++++++
.PURPOSE  Analyze the incoming and get signed number (no E notation)
.RETURNS  How many bytes matched
-----------------*/
{
  char *p, *s ;
    for (s=string; *s == ' '; s++) ;
    p = s ;
    if ((*p == '+') || (*p == '-')) p++ ;
    p += match_unum(p, value) ;
    if (*s == '-') *value = -*value ;
    return(p-string);
}
static int get_center (char *string, double center[2])
/*++++++++++++++++
.PURPOSE  Interpret a string for RA + DEC
.RETURNS  Number of bytes matched
-----------------*/
{
  double v, factor ;
  char *p, *s ;

    factor = _factor2_ ;
    center[0] = center[1] = 0 ;
    p = string ; while (*p == ' ') p++ ;
    p += match_unum(p, center) ;
    while (*p == ' ') p++ ;
    if (*p == ',')      /* For those HEASARC guys ! */
        for (++p; *p == ' '; p++) ;
    if ((*p == '+') || (*p == '-')) ;
    else {              /* Sexagesimal */
        while ((*p == ':') || isdigit(*p)) {
            _factor2_ /= 60.;
            if (*p == ':') p++ ;
            p += match_unum(p, &v);
            while (*p == ' ') p++ ;
            center[0] += v ;
        }
        center[0] *= 15. ;      /* Right Ascension */
    }
    s = p ;
    if ((*s == '+') || (*s == '-')) p++ ;
    _factor2_ = 1 ;
    while ((*p == ':') || isdigit(*p)) {
        if (*p == ':') p++ ;
        p += match_unum(p, &v);
        while (*p == ' ') p++ ;
        center[1] += v ;
        _factor2_ /= 60. ;
    }
    if (*s == '-') center[1] = -center[1] ;

    _factor2_ = factor ;         /* Restore */
    return(p-string) ;
}
